#!ksh
#
# makedist.cmd - korn shell script to make the binary distribution of RLaB
# 
# usage: makedist [options]
#
# makedist recognizes the following options;
#
#	-d dir		use dir as the install top level dir
#
#	-v		print the command line for each subprocess executed
#
################################################################################
#
# use the korn shell for command processor
#
COMSPEC='d:/os2/bin/ksh.exe' ;export COMSPEC

#
# initialize
#
ECHO=echo
LOC_DIR=`pwd`
TGT_DIR=`cd ../../../../RLaB; pwd; cd $LOC_DIR`
vOPT=1

#
# define useful functions
#

check_directory()
{
	test -d "$1"
	case $? in
		0) ;;
		*)	$ECHO "directory: $1 does not exist, or is not a directory" 1>&2
			exit $? ;;
	esac
}

check_file()
{
	test -f "$1"
	case $? in
		0) ;;
		*)	$ECHO "file: $1 does not exist, or is not a regular file" 1>&2
			exit $? ;;
	esac
}

execute_process()
{
	case $vOPT in
		0) $ECHO "$1" 1>&2 ;;
	esac
	eval "$1"
	rc=$?
	case $rc in
		0) ;;
		*) exit $rc ;;
	esac
}

flip_only_files()
{
	if [ X"$1" != X ]; then
		FLIPLIST=
		for filename in $1
			do
				if [ -f $filename ]; then
					FLIPLIST="$FLIPLIST $filename"
				fi
			done
		if [ X"$FLIPLIST" != X ]; then
			execute_process "flip -m -v $FLIPLIST"
		fi
	fi
}

# install trap handler

trap "exit \$rc" 0

# parse command line options

shift 1

set -- `getopt "d:v" $*`

case $? in
	0) ;;
	*) exit 1 ;;
esac

while
	test X$1 != X--
do

	case "$1" in

	-d)
		check_directory "$2"
		TGT_DIR="$2"
		shift 2
		;;

	-v)
		vOPT=0
		shift
		;;

	*)
		$ECHO "warning: option $1 was ignored" 1>&2
		shift
		;;

	esac

done

#
# build the required directory structure in $TGT_DIR
#
echo building RLaB distribution in $TGT_DIR

execute_process "cd $TGT_DIR"
execute_process "mkdir doc"
execute_process "mkdir examples"
execute_process "mkdir help"
execute_process "mkdir misc"
execute_process "mkdir misc/OS2"
execute_process "mkdir rlib"
execute_process "mkdir tmp"
execute_process "mkdir toolbox"

#
# copy the RLaB general distribution files
#
echo copying the RLaB general distribution files to $TGT_DIR

execute_process "cd $LOC_DIR"
execute_process "cp -p ../../examples/* $TGT_DIR/examples/"
execute_process "cp -p ../../help/* $TGT_DIR/help/"
execute_process "cp -p ../../misc/gnuplothelp/* $TGT_DIR/help/"
execute_process "cp -p ../../rlib/* $TGT_DIR/rlib/"
execute_process "cp -p ../../toolbox/* $TGT_DIR/toolbox/"

#
# copy the top level text files
#
echo copying the top level files to $TGT_DIR

execute_process "cp -p ../../ACKNOWLEDGMENT $TGT_DIR"
execute_process "cp -p ../../BUGS $TGT_DIR"
execute_process "cp -p ../../COPYING $TGT_DIR"
execute_process "cp -p ../../PROBLEMS $TGT_DIR"
execute_process "cp -p ../../README $TGT_DIR"
execute_process "cp -p ../../TODO $TGT_DIR"
execute_process "cp -p ../../WHATS.NEW $TGT_DIR"
execute_process "cp -p ../../.rlab $TGT_DIR"
execute_process "cp -p ../../test $TGT_DIR"
execute_process "cp -p ../../test.getline $TGT_DIR"
execute_process "cp -p ../../test.input $TGT_DIR"
execute_process "cp -p README.OS2 $TGT_DIR"
execute_process "cp -p install-wps.cmd $TGT_DIR"

#
# copy the RLaB general distribution documentation files
#
echo copying RLaB general distribution documentation files $TGT_DIR/doc

if [ ! -f ../../doc/rlabp.ps ]; then
  echo "The postscript version of the primer, rlabp.ps, is missing..."
  exit 1
fi

execute_process "cp -p ../../doc/rlabp.tex $TGT_DIR/doc/"
execute_process "cp -p ../../doc/*.ps  $TGT_DIR/doc/"

#
# copy the OS/2 specific documentation files
#
echo copying OS/2 specific files $TGT_DIR/misc/OS2

execute_process "cp -p BUGS.OS2 $TGT_DIR/misc/OS2"
execute_process "cp -p CHANGES.OS2 $TGT_DIR/misc/OS2"
execute_process "cp -p TODO.OS2 $TGT_DIR/misc/OS2"
execute_process "cp -p choparg.cmd $TGT_DIR/misc/OS2"
execute_process "cp -p config.sys $TGT_DIR/misc/OS2"
execute_process "cp -p gnuplot.r $TGT_DIR/rlib/plot.r"
execute_process "cp -p rlab.texi $TGT_DIR/misc/OS2"
execute_process "cp -p rlabgnuplot.cmd $TGT_DIR/misc/OS2"
execute_process "cp -p rlabless.cmd $TGT_DIR/misc/OS2"
execute_process "cp -p rlabrmcat.cmd $TGT_DIR/misc/OS2"

#
# convert all text files to MS-DOG convention
#
echo converting text files to MS-DOG format

execute_process "cd $TGT_DIR"
flip_only_files ".rlab *"
execute_process "cd $TGT_DIR/doc"
flip_only_files "*"
execute_process "cd $TGT_DIR/examples"
flip_only_files "*"
execute_process "cd $TGT_DIR/help"
flip_only_files "*"
execute_process "cd $TGT_DIR/misc"
flip_only_files "*"
execute_process "cd $TGT_DIR/rlib"
flip_only_files "*"
execute_process "cd $TGT_DIR/toolbox"
flip_only_files "*"
execute_process "cd $TGT_DIR/misc/OS2"
flip_only_files "*"

#
# build and copy the RLaB man page
#
echo formatting the RLaB man page

execute_process "cd $LOC_DIR/../../doc"
execute_process "cp -p rlab.1 $TGT_DIR/doc"
execute_process "groff -man rlab.1 > $TGT_DIR/doc/rlab.man"

#
# build the rlab.inf file
#
echo building the RLaB .inf file

execute_process "cd $LOC_DIR"
execute_process "texi2ipf rlab.texi > rlab.ipf"
execute_process "ipfc /INF rlab"
execute_process "mv rlab.INF $TGT_DIR/rlab.inf"

#
# copy the RLaB executable and binary files
#
echo copying the RLaB icon and executable files

if [ ! -f rlab.ico ]; then
  execute_process "uudecode rlab.uue"
fi

execute_process "cp -p rlab.ico $TGT_DIR"
execute_process "cp -p ../../rlab.exe $TGT_DIR"

#
# all done
#
echo all done mahn

exit $rc
