//-------------------------------------------------------------------//
//	Synopsis: RLaB graphics interface

//	The following commands are available in the RLaB graphical
//	interface using GnuPlot as the plotting device.
//	These commands are intended for use with gnuplot-3.6.
//	For details, see separate help file for each command.

//	autoscale ( "axis", I )
//	autotics (  I )
//	axis ( I )
//	cont ( X, Y, Z, KeyTitles, I )
//	copyplot ( I1, I2 )
//	epsplot ( "file_name", "mode", I )
//	grid ( I )
//	hist ( A , N, I )
//	key ( x, y, z, I )
//	keyname ( nr, title, I )
//	label ( "label_text", x, y, z, labelnr, I )	
//	latexplot ( "file_name", "mode", I )
//	linenumbers ( nrs, I )
//	linestyle ( "line_style", I )
//	loglog ( I )
//	loglog2 ( I )
//	multiplot ( rows, columns, I )
//	mx2tics ( freq, I )
//	mxtics ( freq, I )
//	my2tics ( freq, I )
//	mytics ( freq, I )
//	mztics ( freq, I )
//	noautoscale ( "axis", I )
//	noaxis ( I )
//	nogrid ( I )
//	nokey ( I )
//	nolabel ( labelnr, I )
//	nolog ( I )
//	nolog2 ( I )
//	nomultiplot ( I )
//	nomx2tics ( I )
//	nomxtics ( I )
//	nomy2tics ( I )
//	nomytics ( I )
//	nomztics ( I )
//	notime ( I )
//	notitle ( I )
//	nox2tics ( I )
//	noxtics ( I )
//	noy2tics ( I )
//	noytics ( I )
//	noztics ( I )
//	origin ( x-value, y-value, I )
//	pclose ( I )
//	pend ( I )
//	pformat ( "format-string", "axis", I )
//	phelp ( "string", I)
//	plot ( A, KeyTitles, I ) or plot( X, Y, I )
//	plot2 ( A, KeyTitles, I ) or plot( X, Y, I )
//	psplot ( "file_name", "mode", I )
//	printplot ( "printer-name", I )
//	psave ( "file_name", A )
//	psize ( x-value, y-value, I )
//	pssave ( "file_name", X, Y, Z )
//	pstring ( "string", I )
//	range ( x-start, x-end, y-start, y-end, z-start, z-end, I )
//	range2 ( x-start, x-end, y-start, y-end, z-start, z-end, I )
//	replot ( I ) or replot ( A, KeyTitles, I ) or replot ( X, Y, I )
//	replot2 ( I ) or replot2 ( A, KeyTitles, I ) or replot2 ( X, Y, I )
//	semilogx ( I )
//	semilogx2 ( I )
//	semilogy ( I )
//	semilogy2 ( I )
//	setterm ( "term", I )
//	showplot ( I )
//	showpwin ( )
//	splot ( X, Y, Z, KeyTitles, I )
//	time ( I )
//	title ( "title_name", I )
//	x2label ( "string", I )
//	x2tics ( start, incr, end, I )
//	xlabel ( "string", I )
//	xrange ( start, end, I )
//	x2range ( start, end, I )
//	xtics ( start, incr, end, I )
//	ylabel ( "string", I )
//	y2tics ( start, incr, end, I )
//	ylabel ( "string", I )
//	yrange ( start, end, I )
//	y2range ( start, end, I )
//	ytics ( start, incr, end, I )
//	zlabel ( "string", I )
//	zrange ( start, end, I )
//	ztics ( start, incr, end, I )

//-------------------------------------------------------------------//



//
// List to contain plot-object lists.
//

static (p);
p = <<>>;
static (FoS);
FoS = "x1y1";
static (DefPlotWin);
DefPlotWin = 0;

//
// Internal plot-related functions
//

static (plotl, plotm, splotm, plots);
static (replotl, replotm);
static (pobj_Create, pobj_Reset, pobj_Destroy);
static (pobj_TmpFileName, pobj_WriteData, pobj_PlotCmd, pobj_RePlotCmd, pobj_SPlotCmd);
static (pobj_PlotKeyName, pobj_Plot, pobj_SetRm, pobj_Rm, pobj_list);



//
// User interface to plot functionality
//

defplotwin = function ( N )
{
    if (exist(N)) {
	DefPlotWin = N;
	else
	DefPlotWin = 0;
    }
};

plot = function ( data_in, keynames_in, N_in )
{
    FoS = "x1y1";
    plot_cmd ( data_in, keynames_in, N_in );
};

plot2 = function ( data_in, keynames_in, N_in )
{
    FoS = "x1y2";
    plot_cmd ( data_in, keynames_in, N_in );
};

replot = function ( data_in, keynames_in, N_in )
{
    FoS = "x1y1";
    replot_cmd ( data_in, keynames_in, N_in );
};

replot2 = function ( data_in, keynames_in, N_in )
{
    FoS = "x1y2";
    replot_cmd ( data_in, keynames_in, N_in );
};

copyplot = function ( I1_in, I2_in )
{
    if (!exist(I1_in)) { error("One plot-windows must be specified!") }
    if (!exist(I2_in)) {
	I2 = I1;
	I1 = DefPlotWin;
	else
	I1 = I1_in;
	I2 = I2_in;
    }
    if (!exist(p.[I1])) { error("plot-object does not exist!"); }
    if (exist(p.[I2])) {
	pobj_Reset (I2);
	else
	pobj_Create (I2);
    }
    p.[I2].pcmd = p.[I1].pcmd;
    fprintf (p.[I2].prog,p.[I2].pcmd+"\n");
};

plot_cmd = function ( data_in, keynames_in, N_in )
{
    // var = var_in
    if (exist(data_in)) { data = data_in; }
    if (exist(keynames_in)) { keynames = keynames_in; }
    if (exist(N_in)) { N = N_in; }
    
    // Check the existence keynames
    if (exist(keynames)) {
	if (class(keynames) == "num") {
	    if (max(size(keynames)) == 1) {
		N = keynames;
		clear(keynames);
		else
		if (data.nr < data.nc) {
		    data = data';
		}
		if (keynames.nr < keynames.nc) {
		    keynames = keynames';
		}
		data = [data,keynames];
		if (exist(N)) {
		    if (class(N) == "string") {
			keynames = N;
			clear(N);
		    }
		}
	    }
	}
    }
    
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
        
    // Check the existence of p.[N]
    if (max(size(N)) != 1) { error ("splot(): N must be a 1-by-1"); }
    if (exist(p.[N])) {
	pobj_Reset (N);
	else
	pobj_Create (N);
    }
    
    // Set the side to plot on
    p.[N].axis = FoS;

   // Check for multiplot
   if (p.[N].mpp != []) {
	// Delay (RLaB is a bit to fast)
	for(i in 1:1000) { i = i; }
	printf (" Plot number %i/%i done!\n",p.[N].mpp[3],p.[N].mpp[1]*p.[N].mpp[2]);
	if (p.[N].mpp[3] == 1) {
	    fprintf (p.[N].prog, "set multiplot\n");
	    fprintf (p.[N].prog, "set size %s,%s\n",1./p.[N].mpp[2],1./p.[N].mpp[1]);
	}
	N = p.[N].mpp[4];
	fprintf (p.[N].prog, "set origin %s,%s\n",p.[N].mpo[p.[N].mpp[3];1],p.[N].mpo[p.[N].mpp[3];2]);
	p.[N].mpp[3] = p.[N].mpp[3]+1;
   }
   
   // Time to plot
   if (!exist(data)) {
       error ("Data does not exist");
   }
   if (class(data) == "num") {
       pobj_Rm (N);
       plotm (data, N, 1, keynames);
       pobj_Plot (N);
       pobj_SetRm (N);
       else if (class(data) == "string") {
	   plots (data, N);
	   pobj_Plot (N);
	   else if (class(data) == "list") {
	       pobj_Rm (N);
	       plotl (data, N, keynames);
	       pobj_Plot (N);
	       pobj_SetRm (N);	
	       else
	       error ("Invalid data for plot()");
    }   }   }
    
   // Check for multiplot mode
   if (p.[N].mpp != []) {
	if (p.[N].mpp[3] == p.[N].mpp[1]*p.[N].mpp[2]+1) {
	    fprintf (p.[N].prog, "set nomultiplot\n");
	    fprintf (p.[N].prog, "set size %s,%s\n",1,1);
	    fprintf (p.[N].prog, "set origin %s,%s\n",0,0);
# 	    fprintf (p.[N].prog, "set border\n");
	    p.[N].mpp = [];
	    p.[N].mpo = [];
	}
   }
};

//
// User interface to replot functionality
//

replot_cmd = function ( data_in, keynames_in, N_in )
{    
    // var = var_in
    if (exist(data_in)) { data = data_in; }
    if (exist(keynames_in)) { keynames = keynames_in; }
    if (exist(N_in)) { N = N_in; }
    
    // Check the existence keynames
    if (exist(keynames)) {
	if (class(keynames) == "num") {
	    if (max(size(keynames)) == 1) {
		N = keynames;
		clear(keynames);
		else
		if (data.nr < data.nc) {
		    data = data';
		}
		if (keynames.nr < keynames.nc) {
		    keynames = keynames';
		}
		data = [data,keynames];
		if (exist(N)) {
		    if (class(N) == "string") {
			keynames = N;
			clear(N);
		    }
		}
	    }
	}
    }
    
    if (exist(data) && !exist(keynames)) {
	if (max(size(data)) == 1 && class(data) == "num") {
	    N = data;
	    clear(data);
	}
    }
    
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    
    // Check the existence of p.[N]
    if (max(size(N)) != 1) { error ("plot(): N must be a 1-by-1"); }
    if (!exist(p.[N])) {
	error ("plot-object does not exist!");
    }
    
    // Set the side to plot on
    p.[N].axis = FoS;

    // Time to plot
    if (!exist(data)) {
	fprintf (p.[N].prog, "replot\n");
	return 0;
    }
    if (class(data) == "num") {
	p.[N].repnr = p.[N].repnr+1;
	replotm (data, N, 1, keynames);
	pobj_Plot (N);
	pobj_SetRm (N);
	else if (class(data) == "string") {
	    plots (data, N);
	    pobj_Plot (N);
	    fprintf (p.[N].prog, "replot\n");
	    else if (class(data) == "list") {
		p.[N].repnr = p.[N].repnr+1;
		replotl (data, N, keynames);
		pobj_Plot (N);
		pobj_SetRm (N);	
		else
		error ("Invalid data for replot()");
    }   }   }
};


//
// User interface to splot functionality
//

splot = function ( datax_in, datay_in, datazz_in, keynames_in, N_in )
{
    // var = var_in
    if (exist(datax_in)) { datax = datax_in; }
    if (exist(datay_in)) { datay = datay_in; }
    if (exist(datazz_in)) { datazz = datazz_in; }
    if (exist(keynames_in)) { keynames = keynames_in; }
    if (exist(N_in)) { N = N_in; }
    
    // Check what has been sent here
    // One input!
    if (!exist(datay)) {
	datazz = datax;
	datax = 1:datazz.nc;
	datax = datax';
	datay = 1:datazz.nr;
	datay = datay';
	// Two inputs!
	else if (!exist(datazz)) {
	    if (exist(datay)) {
		if (class(datay) == "num") {
		    N = datay;
		    else
		    keynames = datay;
		}
	    }
	    datazz = datax;
	    datax = 1:datazz.nc;
	    datax = datax';
	    datay = 1:datazz.nr;
	    datay = datay';
	    // Three inputs!
	    else if (!exist(keynames)) {
		if (class(datay) != "num") {
		    keynames = datay;
		    N = datazz;
		    datazz = datax;
		    datax = 1:datazz.nc;
		    datax = datax';
		    datay = 1:datazz.nr;
		    datay = datay';
		}
		// Four inputs!
		else if (!exist(N)) {
		    if (class(keynames) == "num") {
			N = keynames;
			clear(keynames);
		    }
    }   }   }   }
    
    // Transpose x- and y-vectors if nessecary
    if (datax.nc != 1 ) {
	if (datax.nr != 1 ) {
	    error ("Invalid data for x-axis (not a vector)");
	    else
	    datax = datax';
	}
    }
    if (datay.nc != 1 ) {
	if (datay.nr != 1 ) {
	    error ("Invalid data for y-axis (not a vector)");
	    else
	    datay = datay';
	}
    }
    
    // Set the default splot-object
    if (!exist(N)) { N = DefPlotWin; }

    // Check the existence of p.[N]
    if (max(size(N)) != 1) { error ("splot(): N must be a 1-by-1"); }
    if (exist(p.[N])) {
	pobj_Reset (N);
	else
	pobj_Create (N);
    }

   // Check for multiplot
   if (p.[N].mpp != []) {
	// Delay (RLaB is a bit to fast)
	for(i in 1:1000) { i = i; }
	printf (" Plot number %i/%i done!\n",p.[N].mpp[3],p.[N].mpp[1]*p.[N].mpp[2]);			
	if (p.[N].mpp[3] == 1) {
	    fprintf (p.[N].prog, "set multiplot\n");
	    fprintf (p.[N].prog, "set size %s,%s\n",1./p.[N].mpp[2],1./p.[N].mpp[1]);
	}
	N = p.[N].mpp[4];
	fprintf (p.[N].prog, "set origin %s,%s\n",p.[N].mpo[p.[N].mpp[3];1],p.[N].mpo[p.[N].mpp[3];2]);
	p.[N].mpp[3] = p.[N].mpp[3]+1;
   }

    // Check how many splots to do
    plotantal = (datazz.nc/datazz.nr)/(datax.nr/datay.nr);
    if (plotantal == int(plotantal)) {
	
	// Time to plot
	if (class(datax) != "num") {
	    error ("Invalid data class for splot() in x-vector (must be num)");
	    else if (class(datay) != "num") {
		error ("Invalid data class for splot() in y-vector (must be num)");
		else if (class(datazz) != "num") {
		    error ("Invalid data class for splot() in x-matrix (must be num)");
		    else
		    fprintf (p.[N].prog, "set view 60, 30, 1, 1\n");
		    fprintf (p.[N].prog, "set surface\n");
		    fprintf (p.[N].prog, "set nocontour\n");
		    pobj_Rm (N);
		    splotm (datax, datay, datazz, N, 1, keynames);
		    pobj_Plot (N);
		    pobj_SetRm (N);
	}   }   }
	else
	error ("Invalid data for splot(), bad x/y-range in z-matrix");
    }
    
   // Check for multiplot mode
   if (p.[N].mpp != []) {
	if (p.[N].mpp[3] == p.[N].mpp[1]*p.[N].mpp[2]+1) {
	    fprintf (p.[N].prog, "set nomultiplot\n");
	    fprintf (p.[N].prog, "set size %s,%s\n",1,1);
	    fprintf (p.[N].prog, "set origin %s,%s\n",0,0);
# 	    fprintf (p.[N].prog, "set border\n");
	    p.[N].mpp = [];
	    p.[N].mpo = [];
	}
   }
};



//
// User interface to splot functionality as 2D-contour
//

cont = function ( datax_in, datay_in, datazz_in, keynames_in, N_in )
{
    // var = var_in
    if (exist(datax_in)) { datax = datax_in; }
    if (exist(datay_in)) { datay = datay_in; }
    if (exist(datazz_in)) { datazz = datazz_in; }
    if (exist(keynames_in)) { keynames = keynames_in; }
    if (exist(N_in)) { N = N_in; }
    
    // Check what has been sent here
    // One input!
    if (!exist(datay)) {
	datazz = datax;
	datax = 1:datazz.nc;
	datax = datax';
	datay = 1:datazz.nr;
	datay = datay';
	// Two inputs!
	else if (!exist(datazz)) {
	    if (exist(datay)) {
		if (class(datay) == "num") {
		    N = datay;
		    else
		    keynames = datay;
		}
	    }
	    datazz = datax;
	    datax = 1:datazz.nc;
	    datax = datax';
	    datay = 1:datazz.nr;
	    datay = datay';
	    // Three inputs!
	    else if (!exist(keynames)) {
		if (class(datay) != "num") {
		    keynames = datay;
		    N = datazz;
		    datazz = datax;
		    datax = 1:datazz.nc;
		    datax = datax';
		    datay = 1:datazz.nr;
		    datay = datay';
		}
		// Four inputs!
		else if (!exist(N)) {
		    if (class(keynames) == "num") {
			N = keynames;
			clear(keynames);
		    }
    }   }   }   }
    
    // Transpose x- and y-vectors if nessecary
    if (datax.nc != 1 ) {
	if (datax.nr != 1 ) {
	    error ("Invalid data for x-axis (not a vector)");
	    else
	    datax = datax';
	}
    }
    if (datay.nc != 1 ) {
	if (datay.nr != 1 ) {
	    error ("Invalid data for y-axis (not a vector)");
	    else
	    datay = datay';
	}
    }
    
    // Set the default splot-object
    if (!exist(N)) { N = DefPlotWin; }

    // Check the existence of p.[N]
    if (max(size(N)) != 1) { error ("splot(): N must be a 1-by-1"); }
    if (exist(p.[N])) {
	pobj_Reset (N);
	else
	pobj_Create (N);
    }

   // Check for multiplot
   if (p.[N].mpp != []) {
	// Delay (RLaB is a bit to fast)
	for(i in 1:1000) { i = i; }
	if (p.[N].mpp[3] == 1) {
	    fprintf (p.[N].prog, "set multiplot\n");
	    fprintf (p.[N].prog, "set size %s,%s\n",1/p.[N].mpp[2],1/p.[N].mpp[1]);
	}
	N = p.[N].mpp[4];
	fprintf (p.[N].prog, "set origin %s,%s\n",p.[N].mpo[p.[N].mpp[3];1],p.[N].mpo[p.[N].mpp[3];2]);
	p.[N].mpp[3] = p.[N].mpp[3]+1;
   }

    // Check how many splots to do
    plotantal = (datazz.nc/datazz.nr)/(datax.nr/datay.nr);
    if (plotantal == int(plotantal)) {

	// Time to plot
	if (class(datax) != "num") {
	    error ("Invalid data class for splot() in x-vector (must be num)");
	    else if (class(datay) != "num") {
		error ("Invalid data class for splot() in y-vector (must be num)");
		else if (class(datazz) != "num") {
		    error ("Invalid data class for splot() in x-matrix (must be num)");
		    else
		    fprintf (p.[N].prog, "set view 0, 0, 1\n");
		    fprintf (p.[N].prog, "set nosurface\n");
		    fprintf (p.[N].prog, "set contour\n");
		    pobj_Rm (N);
		    splotm (datax, datay, datazz, N, 1, keynames);
		    pobj_Plot (N);
		    pobj_SetRm (N);
	}   }   }
	else
	error ("Invalid data for splot(), bad x/y-range in z-matrix");
    }
    
   // Check for multiplot mode
   if (p.[N].mpp != []) {
	if (p.[N].mpp[3] == p.[N].mpp[1]*p.[N].mpp[2]+1) {
	    fprintf (p.[N].prog, "set nomultiplot\n");
	    fprintf (p.[N].prog, "set size %s,%s\n",1,1);
	    fprintf (p.[N].prog, "set origin %s,%s\n",0,0);
# 	    fprintf (p.[N].prog, "set border\n");
	    p.[N].mpp = [];
	    p.[N].mpo = [];
	}
   }
};



//
// User interface to plot histograms
//

hist = function ( data_in, I, keynames_in, N )
{
    // var = var_in
    if (exist(data_in)) { data = data_in; }
    if (exist(keynames_in)) { keynames = keynames_in; }
    // Set the default number of bins
    if (!exist(I)) { I = 10; }
    // Three inputs
    if (exist(keynames_in)) {
	if (class(keynames) == "num") {
	    N = keynames;
	    clear(keynames);
	}
    }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }

    // sort data
    data = sort(data).val;
    for (col in 1:data.nc) {
	i = 1;
	matrix.[col] = [];
	minval = min(min(data[;col]));
	maxval = max(max(data[;col]));
	stepval = (maxval-minval)/(I);
	matrix.[col][i;1] = minval + (i-1+0.5)*stepval;
	matrix.[col][i;2] = 0;
	for (j in 1:data.nr) {
	    if (data[j;col] >= matrix.[col][i;1]-0.5*stepval && data[j;col] <= matrix.[col][i;1]+0.5*stepval) {
		matrix.[col][i;2] = matrix.[col][i;2] + 1;
		else
		if (data[j;col] >= matrix.[col][i;1]+0.5*stepval) {
		    i++;
		    matrix.[col][i;1] = minval + (i-1+0.5)*stepval;
		    matrix.[col][i;2] = 0;
		}
	    }
	}
    }
    minval = min(min(data));
    maxval = max(max(data));
    stepval = (maxval-minval)/(I);
    xrange ((minval-stepval)[1],(maxval+stepval)[1],N);
    ymax = 0;
    for (i in 1:size(matrix)) {
	newymax = max(max(matrix.[i][;2]));
	if (newymax > ymax) {ymax = newymax;}
    }
    yrange (0,int(newymax+1)[1],N);
    if (exist(keynames)) {
	plot (matrix,keynames,N);
	else
	plot (matrix,N);
    }
};

//
// Send string to GNUPLOT
//

pstring = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "%s\n",string);
};



//
// User interface to save a plot ot splot as postscript hardcopy
//

psplot = function ( file, mode_in, N )
{
    // var = var_in
    if (exist(mode_in)) { mode = mode_in; }

    if (!exist(N) && exist(mode)) {
	if (class(mode) == "num") {
	    N = mode;
	}
    }

    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    if (!exist(p.[N])) { error ("No existing plot to make hardcopy from"); }

    // Make hardcopy file
    if (!exist(mode)) { mode = ""; }
    if (class(mode) == "num") { mode = ""; }
    fprintf (p.[N].prog, "set term post default enh %s\n", mode);
    fprintf (p.[N].prog, "set output \"%s\"\n", file);
    fprintf (p.[N].prog, "replot\n");
    
    // Reset to original term type, and replot
    fprintf (p.[N].prog, "set term %s\n", p.[N].term);
    fprintf (p.[N].prog, "replot\n");

    // Delay (RLaB is a bit to fast)
    for(i in 1:50000) { i = i; }
};
postplot = psplot;

epsplot = function ( file, mode_in, N )
{
    // var = var_in
    if (exist(mode_in)) { mode = mode_in; }

    if (!exist(N) && exist(mode)) {
	if (class(mode) == "num") {
	    N = mode;
	}
    }

    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    if (!exist(p.[N])) { error ("No existing plot to make hardcopy from"); }

    // Make hardcopy file
    if (!exist(mode)) { mode = ""; }
    if (class(mode) == "num") { mode = ""; }
    fprintf (p.[N].prog, "set term post epsf enh %s\n", mode);
    fprintf (p.[N].prog, "set output \"%s\"\n", file);
    fprintf (p.[N].prog, "replot\n");
    
    // Reset to original term type, and replot
    fprintf (p.[N].prog, "set term %s\n", p.[N].term);
    fprintf (p.[N].prog, "replot\n");

    // Delay (RLaB is a bit to fast)
    for(i in 1:50000) { i = i; }
};


//
// User interface to save a plot ot splot as LaTeX hardcopy
//

latexplot = function ( file, mode_in, N )
{
    // var = var_in
    if (exist(mode_in)) { mode = mode_in; }
    
    if (!exist(N) && exist(mode)) {
	if (class(mode) == "num") {
	    N = mode;
	}
    }

    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    if (!exist(p.[N])) { error ("No existing plot to make hardcopy from"); }

    // Make hardcopy file
    if (!exist(mode)) { mode = "default"; }
    if (class(mode) == "num") { mode = "default"; }
    fprintf (p.[N].prog, "set term latex %s\n", mode);
    fprintf (p.[N].prog, "set output \"%s\"\n", file);
    fprintf (p.[N].prog, "replot\n");

    // Reset to original term type, and replot
    fprintf (p.[N].prog, "set term %s\n", p.[N].term);
    fprintf (p.[N].prog, "replot\n");
};


//
// User interface to print a plot ot splot
//

printplot = function ( printername_in, N )
{
    // var = var_in
    if (exist(printername_in)) { printername = printername_in; }

    if (exist(printername) && !exist(N)) {
	if (class(printername) == "num") {
	    N = printername;
	    clear(printername);
	}
    }
    
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    
    // Create tmp-file-name
    fn = "/usr/tmp/rlab-tmpf.ps";

    psplot (fn, "", N);
    if (exist(printername)) {
	system ("lp -d" + printername + " " + fn);
	else
	system ("lp " + fn);
    }
    // Delay (RLaB is a bit to fast)
    for(i in 1:10000) { i = i; }
    system ("rm -f " + fn);
};


//
// User interface to save plots as GNUPLOT-datafiles
//

psave = function ( file, data_in )
{
    // var = var_in
    if (exist(data_in)) { data = data_in; }

    // Check file-name
    if (class(file) != "string") {
	error ("Invalid string for file_name");
    }

    // Determine how many lines to save
    nplot = max([1, data.nc - 1]);

    if (nplot > data.nr) {
	printf (" Save %i columns, are you sure [y(es)/n(o)/t(ranspose) (def. no)] ? ", data.nc);
	ans = getline ("stdin");
	if (ans.[1] != "y" && ans.[1] != "t") {
	    return 0;
	}
	if (ans.[1] == "t") {
	    data = data';
	    nplot = max([1, data.nc - 1]);
	}
    }

    // Generate two column matrices for saving
    for (i in 1:nplot) {

	// Write data to file
	if (nplot == 1) {
	    pobj_WriteData (real(data), file);
	    else
	    pobj_WriteData (real(data[;1,i+1]), file);
	}
    }
};



//
// User interface to save splots as GNUPLOT-datafiles
//

pssave = function ( file, datax_in, datay_in, datazz_in)
{
    // var = var_in
    if (exist(datax_in)) { datax = datax_in; }
    if (exist(datay_in)) { datay = datay_in; }
    if (exist(datazz_in)) { datazz = datazz_in; }

    // Determine how many plots to draw
    if (!exist(datazz)) { 
	datazz = datax;
	datax = 1:datazz.nr;
	datax = datax';
	datay = 1:datazz.nc;
	datay = datay';
    }

    if (datax.nc != 1 ) {
	if (datax.nr != 1 ) {
	    error ("Invalid data for x-axis (not a vector)");
	    else
	    datax = datax';
	}
    }
    if (datay.nc != 1 ) {
	if (datay.nr != 1 ) {
	    error ("Invalid data for y-axis (not a vector)");
	    else
	    datay = datay';
	}
    }

    // Create file-name
    if (!exist(file)) { error ("Invalid string for file_name"); }
    if (class(file) != "string") { error ("Invalid string for file_name"); }

    nsplot = (datazz.nr/datazz.nc)/(datax.nr/datay.nr);

    // Generate three column matrices for splot program
    // Write data to new data-file
    for (i in 1:nsplot) {
	newdata = [];
	for (m in 1:datax.nr){
	    rownumber = 0;
	    for (n in 1:datay.nr){
		rownumber = rownumber+1;
		// x-value
		newdata[rownumber;1] = datax[m];
		// y-value
		newdata[rownumber;2] = datay[n];
		// z-value
		newdata[rownumber;3] = datazz[m;n+(i-1)*datax.nr];
	    }
	    writem (file,real(newdata[;1,2,3]));
	    writem (file,"");
	}
	close (file);
    }
};



//
// Make multiplots
//

multiplot = function ( row, col, N )
{
   // Set the default plot-object
   if (exist(row) && !exist(col)) { N = row; }
   if (!exist(N)) { N = DefPlotWin; }
   // Check the existence of p.[N]
   if (!exist(p.[N])) { pobj_Create (N); }
   // Check if to set multiplot or to do a multiple plot.
   if (!exist(col)) {
	fprintf (p.[N].prog, "set multiplot\n");
	else 
	p.[N].mpp = [row,col,1,N];
	p.[N].mpo = [];
	printf (" Make %i plots!\n",p.[N].mpp[1]*p.[N].mpp[2]);
	k = 1;
	for (i in 1:row) {
	    for (j in 1:col) {
		p.[N].mpo[k;1] = (j-1)/col;
		p.[N].mpo[k;2] = (row-i)/row;
		k = k+1;
	    }
	}
   }
};




//
// End multiplot mode
//

nomultiplot = function ( N )
{
   // Set the default plot-object
   if (!exist(N)) { N = DefPlotWin; }
   // Check the existence of p.[N]
   if (!exist(p.[N])) { pobj_Create (N); }
   fprintf (p.[N].prog, "set size %s,%s\n",1,1);
   fprintf (p.[N].prog, "set origin %s,%s\n",0,0);
   fprintf (p.[N].prog, "set nomultiplot\n");
   p.[N].mpp = [];
   p.[N].mpo = [];
};


//
// User interface to close a GNUPLOT-window
//

pclose = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    if (N == "all") {
	plwins = members(p);
	for (i in 1:plwins.nc) {
	    N = eval(plwins[i]);
	    if (exist(p.[N])) {
		close (p.[N].rmf);
		close (p.[N].prog);
		pobj_Destroy (N);
	    }
	}
	N = "all";
	else
	if (!exist(p.[N])) { return -1; }
	close (p.[N].rmf);
	close (p.[N].prog);
	pobj_Destroy (N);
    }
    return N;
};

pend = function ()
{
    plwins = members(p);
    for (i in 1:plwins.nc) {
	N = eval(plwins[i]);
	if (exist(p.[N])) {
	    close (p.[N].rmf);
	    close (p.[N].prog);
	    pobj_Destroy (N);
	}
    }
};



//
// Set the terminal type for the I-th GNUPLOT process
//

setterm = function ( TERM, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    p.[N].term = TERM;

    // Now send the "set term" command to GNUPLOT
    fprintf (p.[N].prog, "set term %s\n", TERM);
};



//
// Print out the plot list
//

showplot = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    if (!exist(p.[N])) {
	pobj_Create (N);
    }

    printf ("\tPlot List %i\n", N);
    printf ("\t\tTerm:\t\t\t%s\n", p.[N].term);
    printf ("\t\tOutput:\t\t\t%s\n", p.[N].output);
    printf ("\t\tTmp Files:\t\t%s\n", p.[N].files);
    //	printf ("\t\tHardCopy:\t\t%s\n", p.[N].hrd);
    //	printf ("\t\tHardCopy Files:\t\t%s\n", p.[N].hrdf);
};

showpwin = function ()
{
    for ( i in 0:99) {
	if (exist(p.[i])) {
	    printf ("\tPlot window number %i exists!\n", i);
	}
    }
};



//
// Here comes a bunch of functions that really is not nessecary.
// They could be done with string("set key") etc, but this is more convinient.
//

phelp = function ( string_in, N )
{
    // Set the default plot-object
    if (exist(string_in) && !exist(N)) {
	if (class(string_in) == "num") {
	    N = string_in;
	    string = "";
	    else
	    string = string_in;
	}
    }
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (!exist(string)) { string = ""; }
    fprintf (p.[N].prog, "help %s\n",string);
};

xrange = function ( start, end, N )
{
    // Set the default plot-object
    if (exist(start) && !exist(end)) { N = start; }
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (!exist(end)) {
	fprintf (p.[N].prog, "set xrange [] writeback\n");
	else
	fprintf (p.[N].prog, "set xrange [%s:%s]\n",start,end);
    }
};

yrange = function ( start, end, N )
{
    // Set the default plot-object
    if (exist(start) && !exist(end)) { N = start; }
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (!exist(end)) {
	fprintf (p.[N].prog, "set yrange [] writeback\n");
	else
	fprintf (p.[N].prog, "set yrange [%s:%s]\n",start,end);
    }
};

zrange = function ( start, end, N )
{
    // Set the default plot-object
    if (exist(start) && !exist(end)) { N = start; }
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (!exist(end)) {
	fprintf (p.[N].prog, "set zrange [] writeback\n");
	else
	fprintf (p.[N].prog, "set zrange [%s:%s]\n",start,end);
    }
};

range = function ( startx, endx, starty, endy, startz, endz, N )
{
    // Check what has been sent
    if (exist(startz) && !exist(endz)) { N = startz; }
    if (exist(starty) && !exist(endy)) { N = starty; }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set xrange [%s:%s]\n",startx,endx);
    if (exist(endy)) {
	fprintf (p.[N].prog, "set yrange [%s:%s]\n",starty,endy);
    }
    if (exist(endz)) {
	fprintf (p.[N].prog, "set zrange [%s:%s]\n",startz,endz);
    }
};
x2range = function ( start, end, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set x2range [%s:%s]\n",start,end);
};

y2range = function ( start, end, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set y2range [%s:%s]\n",start,end);
};

range2 = function ( startx, endx, starty, endy, startz, endz, N )
{
    // Check what has been sent
    if (exist(startz) && !exist(endz)) { N = startz; }
    if (exist(starty) && !exist(endy)) { N = starty; }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set x2range [%s:%s]\n",startx,endx);
    if (exist(endy)) {
	fprintf (p.[N].prog, "set y2range [%s:%s]\n",starty,endy);
    }
    if (exist(endz)) {
	fprintf (p.[N].prog, "set zrange [%s:%s]\n",startz,endz);
    }
};

autotics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set xtics\n");
    fprintf (p.[N].prog, "set ytics\n");
    fprintf (p.[N].prog, "set ztics\n");
    fprintf (p.[N].prog, "set nox2tics\n");
    fprintf (p.[N].prog, "set noy2tics\n");
    fprintf (p.[N].prog, "set nomxtics\n");
    fprintf (p.[N].prog, "set nomytics\n");
    fprintf (p.[N].prog, "set nomztics\n");
    fprintf (p.[N].prog, "set nomx2tics\n");
    fprintf (p.[N].prog, "set nomy2tics\n");
};

xtics = function ( start, incr, end, N )
{
    if (!exist(start)) { start = 0; }
    if (!exist(end)) {
	if (max(size(start)) > 1) {
	    if (!exist(incr)) { incr = 0; }
	    if (!exist(p.[incr])) { pobj_Create (incr); }
	    string = "set xtics (" + num2str(start[1])[1];
	    for (i in 2:max(size(start))) {
		if(class(i) == "string"){
		    string = string + "," + i;
		    else
		    string = string + "," + num2str(start[i])[1];
		}
	    }
	    string = string + ")\n";
	    fprintf (p.[incr].prog, string);
	    else
	    if (!exist(p.[start])) { pobj_Create (start); }
	    fprintf (p.[start].prog, "set xtics\n");
	}
	else
	// Set the default plot-object
	if (!exist(N)) { N = DefPlotWin; }
	if (!exist(p.[N])) { pobj_Create (N); }
	fprintf (p.[N].prog, "set xtics %s,%s,%s\n",start,incr,end);
    }
};

noxtics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set noxtics\n");
};

x2tics = function ( start, incr, end, N )
{
    if (!exist(start)) { start = 0; }
    if (!exist(end)) {
	if (max(size(start)) > 1) {
	    if (!exist(incr)) { incr = 0; }
	    if (!exist(p.[incr])) { pobj_Create (incr); }
	    string = "set x2tics (" + num2str(start[1])[1];
	    for (i in 2:max(size(start))) {
		string = string + "," + num2str(start[i])[1];
	    }
	    string = string + ")\n";
	    fprintf (p.[incr].prog, string);
	    else
	    if (!exist(p.[start])) { pobj_Create (start); }
	    fprintf (p.[start].prog, "set x2tics\n");
	}
	else
	// Set the default plot-object
	if (!exist(N)) { N = DefPlotWin; }
	if (!exist(p.[N])) { pobj_Create (N); }
	fprintf (p.[N].prog, "set x2tics %s,%s,%s\n",start,incr,end);
    }
};

nox2tics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nox2tics\n");
};

ytics = function ( start, incr, end, N )
{
    if (!exist(start)) { start = 0; }
    if (!exist(end)) {
	if (max(size(start)) > 1) {
	    if (!exist(incr)) { incr = 0; }
	    if (!exist(p.[incr])) { pobj_Create (incr); }
	    string = "set ytics (" + num2str(start[1])[1];
	    for (i in 2:max(size(start))) {
		string = string + "," + num2str(start[i])[1];
	    }
	    string = string + ")\n";
	    fprintf (p.[incr].prog, string);
	    else
	    if (!exist(p.[start])) { pobj_Create (start); }
	    fprintf (p.[start].prog, "set ytics\n");
	}
	else
	// Set the default plot-object
	if (!exist(N)) { N = DefPlotWin; }
	if (!exist(p.[N])) { pobj_Create (N); }
	fprintf (p.[N].prog, "set ytics %s,%s,%s\n",start,incr,end);
    }
};

noytics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set noytics\n");
};

y2tics = function ( start, incr, end, N )
{
    if (!exist(start)) { start = 0; }
    if (!exist(end)) {
	if (max(size(start)) > 1) {
	    if (!exist(incr)) { incr = 0; }
	    if (!exist(p.[incr])) { pobj_Create (incr); }
	    string = "set y2tics (" + num2str(start[1])[1];
	    for (i in 2:max(size(start))) {
		string = string + "," + num2str(start[i])[1];
	    }
	    string = string + ")\n";
	    fprintf (p.[incr].prog, string);
	    else
	    if (!exist(p.[start])) { pobj_Create (start); }
	    fprintf (p.[start].prog, "set y2tics\n");
	}
	else
	// Set the default plot-object
	if (!exist(N)) { N = DefPlotWin; }
	if (!exist(p.[N])) { pobj_Create (N); }
	fprintf (p.[N].prog, "set y2tics %s,%s,%s\n",start,incr,end);
    }
};

noy2tics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set noy2tics\n");
};

ztics = function ( start, incr, end, N )
{
    if (!exist(start)) { start = 0; }
    if (!exist(end)) {
	if (max(size(start)) > 1) {
	    if (!exist(incr)) { incr = 0; }
	    if (!exist(p.[incr])) { pobj_Create (incr); }
	    string = "set ztics (" + num2str(start[1])[1];
	    for (i in 2:max(size(start))) {
		string = string + "," + num2str(start[i])[1];
	    }
	    string = string + " nomirror)\n";
	    fprintf (p.[incr].prog, string);
	    else
	    if (!exist(p.[start])) { pobj_Create (start); }
	    fprintf (p.[start].prog, "set ztics nomirror\n");
	}
	else
	// Set the default plot-object
	if (!exist(N)) { N = DefPlotWin; }
	if (!exist(p.[N])) { pobj_Create (N); }
	fprintf (p.[N].prog, "set ztics %s,%s,%s nomirror\n",start,incr,end);
    }
};

noztics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set noztics\n");
};

mxtics = function ( freq, N )
{
    if (!exist(freq)) { error("No value given!"); }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set mxtics %s\n",freq);
};

nomxtics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nomxtics\n");
};

mx2tics = function ( freq, N )
{
    if (!exist(freq)) { error("No value given!"); }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set mx2tics %s\n",freq);
};

nomx2tics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nomx2tics\n");
};

mytics = function ( freq, N )
{
    if (!exist(freq)) { error("No value given!"); }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set mytics %s\n",freq);
};

nomytics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nomytics\n");
};

my2tics = function ( freq, N )
{
    if (!exist(freq)) { error("No value given!"); }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set my2tics %s\n",freq);
};

nomy2tics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nomy2tics\n");
};

mztics = function ( freq, N )
{
    if (!exist(freq)) { error("No value given!"); }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set mztics %s\n",freq);
};

nomztics = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nomztics\n");
};

loglog = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set logscale xy\n");
};

semilogx = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nologscale xy\n");
    fprintf (p.[N].prog, "set logscale x\n");
};

semilogy = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nologscale xy\n");
    fprintf (p.[N].prog, "set logscale y\n");
};

nolog = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nologscale xy\n");
};

loglog2 = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set logscale x2\n");
    fprintf (p.[N].prog, "set logscale y2\n");
};

semilogx2 = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nologscale y2\n");
    fprintf (p.[N].prog, "set logscale x2\n");
};

semilogy2 = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nologscale x2\n");
    fprintf (p.[N].prog, "set logscale y2\n");
};

nolog2 = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nologscale x2\n");
    fprintf (p.[N].prog, "set nologscale y2\n");
};

autoscale = function ( axis_in, N )
{
    // var = var_in
    if (exist(axis_in)) { axis = axis_in; }

    if (!exist(axis) && !exist(N)) {
	N = DefPlotWin;
	axis = ["xy","x2","y2"];
    }
    if (exist(axis) && !exist(N)) {
	if (class(axis) == "num") {
	    N = axis;
	    axis = ["xy","x2","y2"];
	}
    }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    
    if (class(axis) != "string") {
	error ("Must be of class string");
    }
    for (i in 1:max(size(axis))) {
	fprintf (p.[N].prog, "set autoscale %s\n",axis[i]);
    }
};

noautoscale = function ( axis_in, N )
{
    // var = var_in
    if (exist(axis_in)) { axis = axis_in; }
        
    if (!exist(axis) && !exist(N)) {
	N = DefPlotWin;
	axis = ["xy","x2","y2"];
    }
    if (exist(axis) && !exist(N)) {
	if (class(axis) == "num") {
	    N = axis;
	    axis = ["xy","x2","y2"];
	}
    }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }

    if (class(axis) != "string") {
	error ("Must be of class string");
    }
    for (i in 1:max(size(axis))) {
	fprintf (p.[N].prog, "set noautoscale %s\n",axis[i]);
    }
};

psize = function ( x_size, y_size, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    
    fprintf (p.[N].prog, "set size %s,%s\n",x_size,y_size);
};

origin = function ( x_origin, y_origin, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    
    fprintf (p.[N].prog, "set origin %s,%s\n",x_origin,y_origin);
};

xlabel = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set xlabel \"%s\"\n",string);
};

ylabel = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set ylabel \"%s\"\n",string);
};

zlabel = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set zlabel \"%s\"\n",string);
};

x2label = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set x2label \"%s\"\n",string);
};

y2label = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set y2label \"%s\"\n",string);
};

title = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set title \"%s\"\n",string);
};
ptitle = title;

notitle = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set title\n");
};

linestyle = function ( string, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    p.[N].linetype = string;
};

linenumbers = function ( nrs, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    p.[N].lnr = nrs;
};

pformat = function ( string1_in, string2_in, N )
{
    // var = var_in
    if (exist(string1_in)) { string1 = string1_in; }
    if (exist(string2_in)) { string2 = string2_in; }

    if (!exist(string2)) {
	string2 = "xy";
    }
    if (exist(string2) && !exist(N)) {
	if (class(string2) == "num") {
	    N = string2;
	    string2 = "xy";
	}
    }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set format %s \"%s\"\n",string2,string1);
};

#keyname = function ( number_in, keyname_in, N )
#{
#    // var = var_in
#    if (exist(number_in)) { number = number_in; }
#    if (exist(keyname_in)) { keyname = keyname_in; }
#    
#    if (!exist(number)) {
#	error("Wrong number of inputs!");
#    }
#    if (!exist(keyname_in)) {
#	keyname = number;
#    }
#    if (!exist(N)) {
#	if (class(number_in) == "string") {
#	    number = p.[N].repnr;
#	    N = keyname;
#	    keyname = number;
#	}
#    }
#    // Set the default plot-object
#    if (!exist(N)) { N = DefPlotWin; }
#    // Check the existence of p.[N]
#    if (!exist(p.[N])) { pobj_Create (N); }
#    p.[N].keyname[number] = "keyname";
#};


key = function ( x_in, y, z, N )
{
    // var = var_in
    if (exist(x_in)) { x = x_in; }

    if (exist(x) && !exist(y)) {
	if (class(x) == "num") {
	    N = x;
	    else
	    just = x;
	}
    }
    if (exist(x) && exist(y)) {
	if (class(x) == "string") {
	    just = x;
	    N = y;
	}
    }
    if (!exist(z)) { z = 0; }
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (exist(just)) {
	fprintf (p.[N].prog, "set key %s\n",just);
	else
	if (exist(y)) {
	    fprintf (p.[N].prog, "set key %s,%s,%s\n",x,y,z);
	    else
	    fprintf (p.[N].prog, "set key\n");
	}
    }
};

nokey = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nokey\n");
};

grid = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set grid\n");
};

nogrid = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set nogrid\n");
};

axis = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set xzeroaxis\n");
    fprintf (p.[N].prog, "set yzeroaxis\n");
};

noaxis = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set noxzeroaxis\n");
    fprintf (p.[N].prog, "set noyzeroaxis\n");
};

label = function ( label, x, y, z, labelnr, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (class(label) != "string") {
	error ("The label must be a string!")
    }
    if (exist(labelnr)) {
	fprintf (p.[N].prog, "set label %s \"%s\" at %s,%s\n", label, labelnr , x, y, z);
	else
	fprintf (p.[N].prog, "set label \"%s\" at %s,%s\n", label, x, y, z);
    }
};

nolabel = function ( labelnr, N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    if (exist(labelnr)) {
	fprintf (p.[N].prog, "set nolabel %i\n", labelnr);
	else
	fprintf (p.[N].prog, "set nolabel\n");
    }
};

time = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set time\n");
};

notime = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    // Check the existence of p.[N]
    if (!exist(p.[N])) { pobj_Create (N); }
    fprintf (p.[N].prog, "set notime\n");
};


//====================
// Static Functions	=
//====================

plotl = function ( data, N, keynames )
{
    I = 1;
    for (i in members (data)) {
	if (class(data.[i]) == "num") {
	    // Check for plotting keynames
	    pobj_PlotKeyName (keynames, I, N);
	    plotm (data.[i], N, I);
	    I++;
	    else
	    if (class(data.[i]) == "string") {
		if (strsplt(data.[i])[1] == "w") {
		    string = "";
		    for (j in 6:length(strsplt(data.[i]))) {
			string = string + strsplt(data.[i])[j];
		    }
		    p.[N].linetype = string;
		    else
		    if (strsplt(data.[i])[1] == "t") {
			string = "";
			for (j in 7:length(strsplt(data.[i]))) {
			    string = string + strsplt(data.[i])[j];
			}
			p.[N].keyname = string;
		    }
		}
	    }
	}
    }
};



//
// Set-Up to plot a matrix. Columns 2...N against 1st column.
//

plotm = function ( data_in, N, I, keynames )
{
    // var = var_in
    if (exist(data_in)) { data = data_in; }

    // Determine how many lines to draw
    nplot = max([1, data.nc - 1]);

    if (nplot > data.nr) {
	printf (" Plot %i columns, are you sure [y(es)/n(o)/t(ranspose) (def. no)] ? ", data.nc);
	ans = getline ("stdin");
	if (ans.[1] != "y" && ans.[1] != "t") {
	    return 0;
	}
	if (ans.[1] == "t") {
	    data = data';
	    nplot = max([1, data.nc - 1]);
	}
    }

    // Generate two column matrices for plot program
    p.[N].repnr = 1;
    for (i in 1:nplot) {
	// Create tmp-file-name
	fn = pobj_TmpFileName (N, i, I);

	// Check for plotting keynames
	pobj_PlotKeyName (keynames, i, N );
	
	// Write data to tmp-file
	// Add to plot command
	if (nplot == 1) {
	    pobj_WriteData (real(data), fn);
	    pobj_PlotCmd (N, fn);
	    else
	    pobj_WriteData (real(data[;1,i+1]), fn);
	    pobj_PlotCmd (N, fn);
	}
    }
};


replotl = function ( data, N, keynames )
{
    I = 1;
    for (i in members (data)) {
	if (class(data.[i]) == "num") {
	    // Check for plotting keynames
	    pobj_PlotKeyName (keynames, I, N);
	    replotm (data.[i], N, I);
	    I++;
	}
    }
};


replotm = function ( data_in, N, I, keynames )
{
    // var = var_in
    if (exist(data_in)) { data = data_in; }
    
    // Determine how many lines to draw
    nplot = max([1, data.nc - 1]);

    if (nplot > data.nr) {
	printf (" Plot %i columns, are you sure [y(es)/n(o)/t(ranspose) (def. no)] ? ", data.nc);
	ans = getline ("stdin");
	if (ans.[1] != "y" && ans.[1] != "t") {
	    return 0;
	}
	if (ans.[1] == "t") {
	    data = data';
	    nplot = max([1, data.nc - 1]);
	}
    }

    // Generate two column matrices for plot program
    for (i in 1:nplot) {
	// Create tmp-file-name
	fn = pobj_TmpFileName (N, i, I);

	// Check for plotting keynames
	pobj_PlotKeyName (keynames, i, N);

	// Write data to tmp-file
	// Add to plot command
	if (nplot == 1) {
	    pobj_WriteData (real(data), fn);
	    pobj_RePlotCmd (N, fn);
	    else
	    pobj_WriteData (real(data[;1,i+1]), fn);
	    pobj_RePlotCmd (N, fn);
	}
    }
};



//
// Set-Up to splot a matrix. 
//

splotm = function ( datax, datay, datazz, N, I, keynames )
{
    // Determine how many plots to draw
    nsplot = (datazz.nc/datazz.nr)/(datax.nr/datay.nr);

    // Generate three column matrices for splot program
    p.[N].repnr = 1;
    // Write data to new data-file
    // Create tmp-file-name
    for (i in 1:nsplot) {
	fn = pobj_TmpFileName (N, i, I);

	// Check for plotting keynames
	pobj_PlotKeyName (keynames, I, N);
	newdata = [];
	for (m in 1:datax.nr){
	    rownumber = 0;
	    for (n in 1:datay.nr){
		rownumber = rownumber+1;
		// x-value
		newdata[rownumber;1] = datax[m];
		// y-value
		newdata[rownumber;2] = datay[n];
		// z-value
		newdata[rownumber;3] = datazz[m;n+(i-1)*datax.nr];
	    }
	    writem (fn,real(newdata[;1,2,3]));
	    writem (fn,"");
	}
	close (fn);
	// Add to splot command
	pobj_SPlotCmd (N, fn);
    }
};



//
// Form a plain string to send to GNUPLOT as 
// command
//

plots = function ( data, N )
{
    // Send the string to GNUPLOT
    p.[N].pcmd = data + "\n";
};



//
// Create a plot-object.
//

pobj_Create = function ( N )
{
    // The tmp files to plot
    plist.files = "";
    // Where the plot command will go
    plist.pcmd = "";
    
    // Which y-axis
    plist.axis = FoS;
    // Which linestyle
    plist.linetype = "lines";
    // Which keykeyname
    plist.keyname = "";
    #for (i in 1:20) { plist.keyname[i] = "C-" + num2str(i); }
    // Number of plotted curves
    plist.pnr = 0;
    plist.lnr = [1,2,3,4,5,6,7,8];
    // Number of replots
    plist.repnr = 1;
    // Some parameters to control the multiplot functions
    plist.mpp = [];
    plist.mpo = [];
    
    // Init string for plotting program
    plist.init = "set nogrid\nset mxtics 5\nset mytics 5\nset data style lines\nset nokey\nset noxzeroaxis\nset noyzeroaxis\n";

    // The program that draws the plot(s)
    sprintf (plist.prog, "|gnuplot #%i", N);

    // To remove tmp-files

    sprintf (plist.rmf, "|rm -f `cat` #%i", N);

    // Keep track of terminal type, and output
    plist.term = "X11";
    plist.output = "stdout";

    // Copy the local list into the static plot-object collection
    p.[N] = plist;
    fprintf (p.[N].prog, "%s", p.[N].init);
    fprintf (p.[N].prog, "set term %s\n", p.[N].term);
};



//
// Reset a plot object to plot new data
//

pobj_Reset = function ( N )
{

    // Close any existing tmp-files that belong
    // to plot-object N. Reset the file-name list.
    pobj_Rm (N);
    p.[N].files = "";
    p.[N].pcmd = "";
    p.[N].pnr = 0;
};



//
// Destroy a plot-object
//

pobj_Destroy = function ( N )
{
    if (exist(p.[N])) {
	clear (p.[N]);	
    }
};



//
// Create a tmp-file name
//

pobj_TmpFileName = function ( N, i, j )
{
    // N = plot window
    // i = curve number
    // p.[N].repnr = number of replots
    // j = list number
    sprintf (tmp, "/usr/tmp/rlab-tmpf-%i-%i-%i-%i", N, i, p.[N].repnr, j);
    p.[N].files = p.[N].files + " " + tmp;
    return tmp;
};



//
// Add data-file (tmp-file) to plot object list of stuff
// to be plotted.
//

pobj_WriteData = function ( m, file )
{
    writem (file, m);
    close (file);
};



//
// Create the command(s) to plot all the data in the plot object.
//

pobj_PlotCmd = function ( N, fn )
{
    if(p.[N].pnr == length(p.[N].lnr)){
	p.[N].pnr = 1;
	else
	p.[N].pnr = p.[N].pnr + 1;
    }
    if(p.[N].pcmd == "") {
	sprintf(tmp, "plot %s,'%s' title '%s' with %s %s", p.[N].axis, fn, p.[N].keyname, p.[N].linetype, p.[N].lnr[p.[N].pnr]);
	else
	sprintf(tmp, ",%s, '%s' title '%s' with %s %s", p.[N].axis, fn, p.[N].keyname, p.[N].linetype, p.[N].lnr[p.[N].pnr]);
    }
    p.[N].pcmd = p.[N].pcmd + tmp;
    p.[N].keyname = "";
};



//
// Create the command(s) to replot all the data in the plot object.
//

pobj_RePlotCmd = function ( N, fn )
{
    if(p.[N].pnr == length(p.[N].lnr)){
	p.[N].pnr = 1;
	else
	p.[N].pnr = p.[N].pnr + 1;
    }
    if(p.[N].pcmd == "") {
	sprintf(tmp, "replot %s,'%s' title '%s' with %s %s", p.[N].axis, fn, p.[N].keyname, p.[N].linetype, p.[N].lnr[p.[N].pnr]);
	else
	sprintf(tmp, ",%s, '%s' title '%s' with %s %s", p.[N].axis, fn, p.[N].keyname, p.[N].linetype, p.[N].lnr[p.[N].pnr]);
    }
    p.[N].pcmd = p.[N].pcmd + tmp;
    p.[N].keyname = "";
};


//
// Create the command(s) to splot all the data in the splot object.
//

pobj_SPlotCmd = function ( N, fn )
{
    if(p.[N].pnr == length(p.[N].lnr)){
	p.[N].pnr = 1;
	else
	p.[N].pnr = p.[N].pnr + 1;
    }
    if(p.[N].pcmd == "") {
	sprintf(tmp, "splot '%s' title '%s' with %s %s", fn, p.[N].keyname, p.[N].linetype, p.[N].lnr[p.[N].pnr]);
	else
	sprintf(tmp, ", '%s' title '%s' with %s %s", fn, p.[N].keyname, p.[N].linetype, p.[N].lnr[p.[N].pnr]);
    }

    p.[N].pcmd = p.[N].pcmd + tmp;
    p.[N].keyname = "";
};



//
// Check for plotting keynames
//

pobj_PlotKeyName = function (keynames, i, N)
{
    if (exist(keynames)) {
	if (i > max(size(keynames))) {
	    sprintf (p.[N].keyname, "%.4g", i+1);
	    p.[N].keyname = "C-" + p.[N].keyname;
	    else
	    p.[N].keyname = keynames[i];
	}
    }
#    if (!exist(keynames)) {
#	sprintf (p.[N].keyname, "%.4g", i+1);
#	p.[N].keyname = "C-" + p.[N].keyname;
#	else
#	if (i > max(size(keynames))) {
#	    sprintf (p.[N].keyname, "%.4g", i+1);
#	    p.[N].keyname = "C-" + p.[N].keyname;
#	    else
#	    p.[N].keyname = keyname!s[i];
#	}
#    }
};



//
// Force the plot program to create the plot
//

pobj_Plot = function ( N )
{
    //	fprintf ("stdout", "%s\n", p.[N].pcmd);
    fprintf (p.[N].prog, "%s\n", p.[N].pcmd);
};



//
// Setup so the tmp-files can be removed
//

pobj_SetRm = function ( N )
{
    fprintf(p.[N].rmf, "%s", p.[N].files);
};



//
// Remove the tmp-files
//

pobj_Rm = function ( N )
{
    if (length(p.[N].files) != 0) {
	close (p.[N].rmf);
    }
};



//
// Provide access to the plot-list
//

pobj_list = function ( N )
{
    // Set the default plot-object
    if (!exist(N)) { N = DefPlotWin; }
    return p.[N];
};

